use patina::platform;

pub struct SDLInitFlags {
    pub window_width: usize,
    pub window_height: usize,
    pub window_resizable: bool,
}

impl Default for SDLInitFlags {
    fn default() -> Self {
        Self {
            window_width: 800,
            window_height: 600,
            window_resizable: false,
        }
    }
}

pub struct SDLRenderInterface {
    sdl: sdl2::Sdl,
    video: sdl2::VideoSubsystem,
    window: sdl2::video::Window,
}

impl SDLRenderInterface {
    pub fn new(sdlif: SDLInitFlags) -> (Self, SDLEventInterface) {
        let sdl = sdl2::init().unwrap();
        let evt = sdl.event_pump().unwrap();
        let video = sdl.video().unwrap();

        let window = video
            .window(
                "title",
                sdlif.window_width as u32,
                sdlif.window_height as u32,
            )
            .position_centered()
            .build()
            .unwrap();

        (Self { sdl, video, window }, SDLEventInterface { evt, next: None })
    }
}

impl platform::RenderInterface for SDLRenderInterface {}

pub struct SDLEventInterface {
    evt: sdl2::EventPump,
    next: Option<sdl2::event::Event>,
}

impl SDLEventInterface {
    fn translate(&self, sdlevent: sdl2::event::Event) -> Option<platform::event::WindowEvent> {
        match sdlevent {
            sdl2::event::Event::Quit { .. } => Some(platform::event::WindowEvent::CloseRequest),
            _ => None,
        }
    }
}

impl platform::EventInterface for SDLEventInterface {
    fn poll_event(&mut self) -> Option<platform::event::WindowEvent> {
        if let Some(evt) = self.next.take() {
            self.translate(evt)
        } else {
            let sdlevent = self.evt.poll_event()?;
            self.translate(sdlevent)
        }
    }
    fn wait_event(&mut self) {
        if self.next.is_some() { return }
        self.next = Some(self.evt.wait_event());
    }
}

pub struct SDLTextInterface {}

impl platform::TextInterface for SDLTextInterface {}

pub fn init_sdl_platform(
    sdlif: SDLInitFlags,
) -> (SDLRenderInterface, SDLEventInterface, SDLTextInterface) {
    let (sdl, event) = SDLRenderInterface::new(sdlif);
    (sdl, event, SDLTextInterface {})
}
